/* $USAGI: pfkey_v2_msg_add.c,v 1.19 2001/10/22 10:39:14 miyazawa Exp $ */
/*
 * Copyright (C)2001 USAGI/WIDE Project
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * This is a parse routine for a message of SADB_ADD.
 */

#ifdef PFKEY_DEBUG
#define DEBUG 1
#endif

#ifdef MODULE
#include <linux/module.h>
#  ifdef MODVERSIONS
#  include <linux/modversions.h>
#  endif
#endif

#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/random.h>

#include <net/pfkeyv2.h>
#include <net/pfkey.h>
#include <net/sadb.h>

#include "pfkey_v2_msg.h"
#include "sadb_utils.h" /* esptoa authtoa */
			/* sockaddrtoa ultoa */
#define BUFSIZE 64

int sadb_msg_add_parse(struct sock *sk, struct sadb_msg *msg, struct sadb_msg **reply)
{
	int error = 0;
	struct sadb_ext *ext_msgs[SADB_EXT_MAX+1];
	struct sadb_sa *sa;
	struct sadb_address *src;
	struct sadb_address *dst;
	struct sadb_address *pxy;
	struct ipsec_sa sa_entry;

	if (!msg) {
		pr_debug("sadb_msg_add_parse: msg==null\n");
		error = -EINVAL;
		goto err;
	}

	memset(ext_msgs, 0, sizeof(ext_msgs));
	error = sadb_msg_detect_ext(msg, ext_msgs);

	if (ext_msgs[SADB_EXT_SA] &&
		ext_msgs[SADB_EXT_ADDRESS_SRC] &&
		ext_msgs[SADB_EXT_ADDRESS_DST] &&
		(ext_msgs[SADB_EXT_KEY_AUTH] ||
		 ext_msgs[SADB_EXT_KEY_ENCRYPT]))
	{
		sa  = (struct sadb_sa*)ext_msgs[SADB_EXT_SA];

		if ((!sa->sadb_sa_auth && !sa->sadb_sa_encrypt) ||
			sa->sadb_sa_auth > SADB_AALG_MAX ||
			sa->sadb_sa_encrypt > SADB_EALG_MAX ) {
			pr_debug("sadb_msg_add_parse: sa has no transform"
				" or invalid transform value\n");
			error = -EINVAL;
			goto err;
		}

		switch (msg->sadb_msg_satype) {

		case SADB_SATYPE_AH:
			break;
		case SADB_SATYPE_ESP:
			break;
		case SADB_SATYPE_RSVP:
		case SADB_SATYPE_OSPFV2:
		case SADB_SATYPE_RIPV2:
		case SADB_SATYPE_MIP:
		break;

		default:
				
			pr_debug("sadb_msg_add_parse: sa type is ah but no algorithm\n");
			error = -EINVAL;
			goto err;
		}

		if (msg->sadb_msg_satype == SADB_SATYPE_ESP && !(sa->sadb_sa_encrypt)) {
			pr_debug("sadb_msg_add_parse: sa type is esp but no algorithm\n");
			error = -EINVAL;
			goto err;
		}

		ipsec_sa_init(&sa_entry);

		src = (struct sadb_address*)ext_msgs[SADB_EXT_ADDRESS_SRC];
		dst = (struct sadb_address*)ext_msgs[SADB_EXT_ADDRESS_DST];

		sa_entry.ipsec_proto = msg->sadb_msg_satype;
		sa_entry.spi = sa->sadb_sa_spi;


		error = sadb_address_to_sockaddr( src,
				(struct sockaddr*)&(sa_entry.src));
		if (error) {
			pr_debug("sadb_msg_add_parse: src translation failed\n");
			goto err;
		}
		error = sadb_address_to_sockaddr( dst,
				(struct sockaddr*)&(sa_entry.dst));
		if (error) {
			pr_debug("sadb_msg_add_parse: dst translation failed\n");
			goto err;
		}

		if (src->sadb_address_proto == dst->sadb_address_proto) {
			sa_entry.proto = src->sadb_address_proto;
		} else {
			error = -EINVAL;
			goto err;
		} 

		sa_entry.prefixlen_s = src->sadb_address_prefixlen;
		sa_entry.prefixlen_d = dst->sadb_address_prefixlen;

		if ( sa->sadb_sa_auth ) {
			if( (ext_msgs[SADB_EXT_KEY_AUTH]) ) {
				error = sadb_key_to_auth(sa->sadb_sa_auth,
					(struct sadb_key*) ext_msgs[SADB_EXT_KEY_AUTH], &sa_entry);
				if (error)
					goto err;
			} else {
				pr_debug("sadb_msg_add_parse: sa has auth"
					" algo but there is no key for auth\n");
				error = -EINVAL;
				goto err;
			}
		}

		if ( sa->sadb_sa_encrypt ) {
			if( (ext_msgs[SADB_EXT_KEY_ENCRYPT]) ) {
				error = sadb_key_to_esp(sa->sadb_sa_encrypt, 
					(struct sadb_key*) ext_msgs[SADB_EXT_KEY_ENCRYPT], &sa_entry);
				if (error)
					goto err;
			} else {
				pr_debug("sadb_msg_add_parse: sa has esp"
					" algo but there is no key for esp\n");
				error = -EINVAL;
				goto err;
			}
		}

		if (ext_msgs[SADB_EXT_ADDRESS_PROXY]) {
			pxy = (struct sadb_address*)ext_msgs[SADB_EXT_ADDRESS_PROXY];
			error = sadb_address_to_sockaddr( pxy,
					(struct sockaddr*)&(sa_entry.pxy));
			if (error) {
				pr_debug("sadb_msg_add_parse: pxy translation failed\n");
				goto err;
			}
		}

		if (ext_msgs[SADB_EXT_LIFETIME_HARD]) {
			error = sadb_lifetime_to_lifetime((struct sadb_lifetime*)ext_msgs[SADB_EXT_LIFETIME_HARD],
						&(sa_entry.lifetime_h));
			if (error)
				goto err;
		}

		if (ext_msgs[SADB_EXT_LIFETIME_SOFT]) {
			error = sadb_lifetime_to_lifetime((struct sadb_lifetime*)ext_msgs[SADB_EXT_LIFETIME_SOFT],
						&(sa_entry.lifetime_s));
			if (error)
				goto err;
	
		}

		sa_entry.state = SADB_SASTATE_MATURE;

		sadb_append(&sa_entry);	

	} else {
		pr_debug("sadb_msg_add_parse: extensions not enough\n");
		error = -EINVAL;
		goto err;
	}


	/* mask auth key, esp key and proxy address*/
//	ext_msgs[SADB_EXT_KEY_AUTH] = NULL;
//	ext_msgs[SADB_EXT_KEY_ENCRYPT] = NULL;
//	ext_msgs[SADB_EXT_ADDRESS_PROXY] = NULL;
//	error = sadb_msg_reply_build(msg, ext_msgs, reply);

	return error;
err:
	sadb_msg_reply_build2(msg, error, reply);
	return error;
}

